import torch
import os
import pickle
import click
import wandb
from nanoppo.continuous_action_ppo import PPOAgent
from nanoppo.normalizer import Normalizer
from nanoppo.ppo_utils import compute_gae
from nanoppo.environment_manager import EnvironmentManager
from nanoppo.ppo_utils import get_grad_norm

# Memory for PPO
class PPOMemory:
    def __init__(self):
        self.states = []
        self.actions = []
        self.logprobs = []
        self.next_states = [] 
        self.rewards = []
        self.is_terminals = []

    def clear(self):
        del self.states[:]
        del self.actions[:]
        del self.logprobs[:]
        del self.next_states[:]
        del self.rewards[:]
        del self.is_terminals[:]

    def append(self, state, action, logprob, next_state, reward, is_terminal):
        assert(len(self.states) == len(self.actions) == len(self.logprobs) == len(self.next_states) == len(self.rewards) == len(self.is_terminals))
        self.states.append(state)
        self.actions.append(action)
        self.logprobs.append(logprob)
        self.next_states.append(next_state) 
        self.rewards.append(torch.tensor(reward, dtype=torch.float32))  # Convert to tensor here
        self.is_terminals.append(torch.tensor(is_terminal, dtype=torch.float32))  # Convert to tensor here

    def get(self):
        return torch.stack(self.states), torch.stack(self.actions), torch.stack(self.logprobs), torch.stack(self.next_states), torch.stack(self.rewards), torch.stack(self.is_terminals)

def train_agent(env_name, max_episodes=500,
                policy_lr=0.0005, value_lr=0.0005, betas=(0.9, 0.999), 
                n_latent_var=128, gamma=0.99, tau=0.95, K_epochs=4, eps_clip=0.2, 
                vl_coef=0.5, el_coef=0.001,
                max_timesteps=2000, update_timestep=200,
                checkpoint_dir='checkpoints', checkpoint_interval=-1, log_interval=-1, wandb_log=False):
    # Setting up the environment and the agent
    env = EnvironmentManager(env_name).setup_env()
    state_dim = env.observation_space.shape[0]
    action_dim = env.action_space.shape[0]
    print('state_dim', state_dim)
    print('action_dim', action_dim)
    
    checkpoint_path = os.path.join(checkpoint_dir, env_name)
    os.makedirs(checkpoint_path, exist_ok=True)
    print('env', env_name)
    model_file = f"{checkpoint_path}/models.pth"
    metrics_file = f"{checkpoint_path}/metrics.pkl"
    if wandb_log:
        wandb.init(project="nanoPPO", name=env_name, 
                   config={'policy_lr': policy_lr, 'value_lr': value_lr, 'betas': betas, 'gamma': gamma, 
                           'tau': tau, 'K_epochs': K_epochs, 'eps_clip': eps_clip})

    # Initialize a normalizer with the dimensionality of the state
    state_normalizer = Normalizer(state_dim)
    ppo = PPOAgent(state_dim, action_dim, n_latent_var, policy_lr, value_lr, betas, gamma, K_epochs, eps_clip, state_normalizer, 
                   action_low=env.action_space.low, action_high=env.action_space.high, 
                   vl_coef=vl_coef, el_coef=el_coef,
                   wandb_log=wandb_log)
    print(policy_lr, value_lr, betas)
    
    # Load the best weights
    if os.path.exists(model_file):
        metrics = pickle.load(open(metrics_file, 'rb'))
        best_reward = metrics['best_reward']
        start_episode = metrics['episode'] + 1
        ppo.load(model_file)
        print("Loaded best weights!", model_file, metrics_file)
    else:
        best_reward = float('-inf')
        start_episode = 1
    print("best_reward", best_reward)
    print("start_episode", start_episode)
    print("log_interval", log_interval)
    
    ppo_memory = PPOMemory()
    
    # Training loop
    time_step = 0
    avg_length_list = []
    cumulative_reward_list = []  # Initialize cumulative reward
    for episode in range(start_episode, max_episodes + start_episode):
        state, info = env.reset()
        state_normalizer.observe(state)
        state = state_normalizer.normalize(state)
    
        total_reward = 0
        state = torch.FloatTensor(state)
        for t in range(max_timesteps):
            action, log_prob = ppo.policy.act(state)
            action_np = action.detach().numpy()
            next_state, reward, done, truncated, _ = env.step(action_np)
            state_normalizer.observe(next_state)
            next_state = state_normalizer.normalize(next_state)
    
            total_reward += reward
            next_state = torch.FloatTensor(next_state)
            ppo_memory.append(state, action, log_prob, next_state, reward, done)
    
            state = next_state
            time_step += 1
    
            # update if it's time
            if time_step % update_timestep == 0:
                # Get state values for all states
                next_state = torch.FloatTensor(next_state)
                next_value = ppo.policy.get_value(next_state).detach().item()
                values = [ppo.policy.get_value(torch.FloatTensor(state)).item() for state in ppo_memory.states]
                masks = [1 - terminal.item() for terminal in ppo_memory.is_terminals]
                returns = compute_gae(next_value, ppo_memory.rewards, masks, values, gamma=gamma, tau=tau)
                torch_returns = torch.tensor(returns, dtype=torch.float32)
    
                states, actions, log_probs, next_states, rewards, dones = ppo_memory.get()
                ppo.update(states, actions, returns=torch_returns, next_states=next_states, dones=dones)
                ppo_memory.clear()
                time_step = 0
            if done or truncated:
                break
    
        avg_length_list.append(t + 1)
    
        cumulative_reward_list.append(total_reward) 
    
        avg_reward = float(sum(cumulative_reward_list) / len(cumulative_reward_list))
        avg_length = int(sum(avg_length_list) / len(avg_length_list))
        action_mu_grad_norm = get_grad_norm(ppo.policy.action_mu.parameters())
        action_log_std_grad_norm = get_grad_norm(ppo.policy.action_log_std.parameters())
        value_grad_norm = get_grad_norm(ppo.policy.value_layer.parameters())
        # Logging
        if log_interval > 0 and (episode % log_interval == 0):
            sample_length = len(avg_length_list)
            avg_length = int(sum(avg_length_list) / sample_length)
            print(('Episode {} \t samples:{} avg steps: {} \t avg reward: {:.2f} \t best reward: {:.2f} \t'
                   'action_mu_grad_norm: {:.2f} \t action_log_std_grad_norm: {:.2f} \t value_grad_norm: {:.2f}'
                   ).format(
                episode, sample_length, avg_length, avg_reward, best_reward,
                action_mu_grad_norm, action_log_std_grad_norm, value_grad_norm
                ))
            avg_length_list = []
            cumulative_reward_list = []  # Reset cumulative reward after logging

        if checkpoint_interval > 0 and (avg_reward > best_reward):
            print('avg_reward', avg_reward, '> best_reward', best_reward)
            best_reward = avg_reward
            metrics = {'best_reward': best_reward, 'episode':episode}
            pickle.dump(metrics, open(metrics_file, 'wb'))
            ppo.save(model_file)
            print("Saved best weights!", model_file, metrics_file)

        if wandb_log:
            wandb.log({'avg_reward': avg_reward, 'best_reward': best_reward, 'avg_length': avg_length,
                'action_mu_grad_norm': action_mu_grad_norm, 'action_log_std_grad_norm': action_log_std_grad_norm, 'value_grad_norm': value_grad_norm})
    if wandb_log:
        wandb.finish()
    return ppo, model_file, metrics_file
   
@click.command()
@click.option('--env_name', default="PointMass2D-v0", type=click.Choice(["PointMass1D-v0", "PointMass2D-v0", "Pendulum-v1", "MountainCarContinuous-v0"]))
@click.option("--max_episodes", default=100, help="Number of training episodes.")
@click.option("--policy_lr", default=0.0005, help="Learning rate for policy network.")
@click.option("--value_lr", default=0.0005, help="Learning rate for value network.")
@click.option("--vl_coef", default=0.5, help="Value function coefficient.")
@click.option("--checkpoint_dir", default="checkpoints", help="Path to checkpoint.")
@click.option("--checkpoint_interval", default=100, help="Checkpoint interval.")
@click.option("--log_interval", default=10, help="Logging interval.")
@click.option("--wandb_log", is_flag=True, default=False, help="Flag to log results to wandb.")
def cli(env_name, max_episodes, policy_lr, value_lr, vl_coef,
        checkpoint_dir, checkpoint_interval, log_interval, wandb_log):
    ppo, model_file, metrics_file = train_agent(env_name=env_name, max_episodes=max_episodes, policy_lr=policy_lr, value_lr=value_lr,
                                                vl_coef=vl_coef,
                                                checkpoint_dir=checkpoint_dir, 
                                                checkpoint_interval=checkpoint_interval, log_interval=log_interval, 
                                                wandb_log=wandb_log)
    # Load the best weights
    ppo.load(model_file)
    print("Loaded best weights from", model_file)
    metrics = pickle.load(open(metrics_file, 'rb'))
    print("Loaded metrics from", metrics_file)
    best_reward = metrics['best_reward']
    episode = metrics['episode']
    print("best_reward", best_reward, 'episode', episode)
 
if __name__ == '__main__':
    cli()


from sklearn.preprocessing import (
    StandardScaler,
    MinMaxScaler,
    RobustScaler,
    QuantileTransformer,
)
import numpy as np


class StateScaler:
    def __init__(self, env, scale_type, sample_size=10000):
        self.env = env
        self.scale_type = scale_type

        if self.scale_type in ["standard", "minmax", "robust", "quantile"]:
            self.scaler = self._init_scaler(env, sample_size, scale_type)

    def _init_scaler(self, env, sample_size, scale_type):
        samples = [env.observation_space.sample() for _ in range(sample_size)]
        state_space_samples = np.array(
            [
                sample["obs"] if isinstance(sample, dict) else sample
                for sample in samples
            ]
        )

        if scale_type == "standard":
            scaler = StandardScaler()
        elif scale_type == "minmax":
            scaler = MinMaxScaler()
        elif scale_type == "robust":
            scaler = RobustScaler()
        elif scale_type == "quantile":
            scaler = QuantileTransformer()
        else:
            raise ValueError(f"Unknown scale type: {scale_type}")
        scaler.fit(state_space_samples)
        return scaler

    def scale_state(self, state):
        if self.scale_type in ["standard", "minmax", "robust", "quantile"]:
            scaled = self.scaler.transform([state])
            r = scaled[0]  # Return a 1D array instead of 2D
        elif self.scale_type == "env":
            # -1 to 1 scaling
            r = (
                2
                * (state - self.env.observation_space.low)
                / (self.env.observation_space.high - self.env.observation_space.low)
                - 1
            )
        else:
            raise ValueError(f"Unknown scale type: {self.scale_type}")
        return r.astype(np.float32)



import numpy as np

class Normalizer:
    def __init__(self, dim):
        # Mean, standard deviation, and count for each dimension
        self.n = np.zeros(dim, dtype=np.float32)
        self.mean = np.zeros(dim, dtype=np.float32)
        self.mean_diff = np.zeros(dim, dtype=np.float32)
        self.variance = np.zeros(dim, dtype=np.float32)

    def observe(self, x):
        """Update statistics"""
        self.n += 1.0
        last_mean = self.mean.copy()
        self.mean += (x - self.mean) / self.n
        self.mean_diff += (x - last_mean) * (x - self.mean)
        self.variance = (self.mean_diff / self.n).clip(min=1e-2)

    def normalize(self, inputs):
        """Normalize input using running mean and variance"""
        obs_std = np.sqrt(self.variance)
        v = (inputs - self.mean) / obs_std
        return v.astype(np.float32)

    def get_state(self):
        return {
            'n': self.n,
            'mean': self.mean,
            'mean_diff': self.mean_diff,
            'variance': self.variance,
        }
    
    def set_state(self, state):
        self.n = state['n']
        self.mean = state['mean']
        self.mean_diff = state['mean_diff']
        self.variance = state['variance']


import random
from collections import deque
import torch
import numpy as np


class RolloutBuffer:
    def __init__(self, capacity):
        self.capacity = capacity
        self.buffer = deque(maxlen=capacity)

    def push(self, state, action, log_prob, reward, next_state, done):
        self.buffer.append((state, action, log_prob, reward, next_state, done))

    def sample(self, batch_size, device, randomize=True):
        if randomize:
            states, actions, log_probs, rewards, next_states, dones = zip(
                *random.sample(self.buffer, batch_size)
            )
        else:
            states, actions, log_probs, rewards, next_states, dones = zip(
                *list(self.buffer)[-batch_size:]
            )
        states_t = torch.tensor(np.array(states, dtype=np.float32), device=device)
        actions_t = torch.tensor(np.array(actions, dtype=np.float32), device=device)
        log_probs_t = torch.tensor(np.array(log_probs, dtype=np.float32), device=device)
        rewards_t = torch.tensor(np.array(rewards, dtype=np.float32), device=device)
        next_states_t = torch.tensor(np.array(next_states, dtype=np.float32), device=device)
        dones_t = torch.tensor(np.array(dones, dtype=np.float32), device=device)
        return states_t, actions_t, log_probs_t, rewards_t, next_states_t, dones_t

    def __len__(self):
        return len(self.buffer)

    def clear(self):
        self.buffer.clear()



import wandb
import numpy as np

class WandBLogger:
    @staticmethod
    def init(project, name, config):
        wandb.init(project=project, name=name, config=config)

    @staticmethod
    def log(data):
        wandb.log(data)

    @staticmethod
    def finish():
        wandb.finish()

    @staticmethod
    def log_rewards(rewards):
        # Log the rewards during training
        wandb.log(
            {
                "Reward/Min": min(rewards),
                "Reward/Mean": sum(rewards) / len(rewards),
                "Reward/Max": max(rewards),
            }
        )

    @staticmethod
    def log_action_distribution_parameters(action_mean:np.ndarray, action_std:np.ndarray):
        """
        Logs the parameters (mean and standard deviation) of the action distribution using WandB.

        Args:
            action_mean (torch.Tensor): Mean values of the action distribution.
            action_std (torch.Tensor): Standard deviation values of the action distribution.
        """
        log_data = {
            f"ActionDist/Action{i}_Mean": mean_val
            for i, mean_val in enumerate(action_mean.tolist())
        }
        log_data.update(
            {
                f"ActionDist/Action{i}_Std": std_val
                for i, std_val in enumerate(action_std.tolist())
            }
        )
        wandb.log(log_data)



import numpy as np


class RewardScaler:
    """
    This class is designed to normalize the rewards using Welford's online algorithm
    for computing running mean and variance. Normalizing rewards can be beneficial
    for training as it ensures that the scale of the rewards is roughly consistent,
    which can improve the stability and convergence speed of the learning algorithm.

    The formula for Welford's algorithm is defined as:

    M_{k} = M_{k-1} + (x_k - M_{k-1})/k
    S_{k} = S_{k-1} + (x_k - M_{k-1})(x_k - M_{k})

    Where:
    M_k: Running mean after observing k samples.
    S_k: Running sum of squares of differences from the current mean after observing k samples.
    x_k: k-th sample.

    The variance is then computed as:
    variance = S_k / (k-1)
    (for k > 1).
    """

    def __init__(self):
        self.running_mean = 0  # Running mean of rewards.
        self.running_sum_of_square_diffs = (
            0  # Running sum of squares of differences from the current mean.
        )
        self.count = 0  # Count of observed samples.

    def update(self, rewards):
        """
        Update the running mean and variance using Welford's online algorithm.

        Parameters:
        - rewards (list or array): Batch of rewards from recent episodes.
        """
        batch_mean = np.mean(rewards)
        batch_count = len(rewards)
        new_count = self.count + batch_count

        delta = batch_mean - self.running_mean
        new_mean = self.running_mean + delta * batch_count / new_count

        for reward in rewards:
            delta = reward - new_mean
            delta2 = reward - self.running_mean
            self.running_sum_of_square_diffs += delta * delta2

        self.running_mean = new_mean
        self.count = new_count

    def scale_rewards(self, rewards):
        """
        Scale rewards using the running statistics and update the statistics.

        Parameters:
        - rewards (list or array): Batch of rewards from recent episodes.

        Returns:
        - Normalized rewards.
        """
        self.update(rewards)
        variance = (
            self.running_sum_of_square_diffs / (self.count - 1)
            if self.count > 1
            else 1.0
        )
        std = np.sqrt(variance)

        return [(reward - self.running_mean) / (std + 1e-8) for reward in rewards]



import torch
import torch.nn as nn
import torch.optim as optim

# Neural Network for Actor-Critic
import torch.nn as nn
import torch.nn.functional as F
from nanoppo.policy.actor_critic import ActorCritic
import wandb

# PPO Agent
class PPOAgent:
    def __init__(self, state_dim, action_dim, n_latent_var, policy_lr, value_lr, betas, gamma, K_epochs, eps_clip,
                 state_normalizer, action_low, action_high, vl_coef=0.5, el_coef=0.001, wandb_log=False):
        self.gamma = gamma
        self.eps_clip = eps_clip
        self.K_epochs = K_epochs
        self.state_normalizer = state_normalizer
        self.action_low = action_low
        self.action_high = action_high
        self.vl_coef = vl_coef
        self.el_coef = el_coef

        self.policy = ActorCritic(state_dim, action_dim, n_latent_var, action_low, action_high).float()
        self.policy_old = ActorCritic(state_dim, action_dim, n_latent_var, action_low, action_high).float()
        self.policy_old.load_state_dict(self.policy.state_dict())

        # Separate the parameters of the actor and critic networks
        actor_params = list(self.policy.action_mu.parameters()) + list(self.policy.action_log_std.parameters())
        critic_params = list(self.policy.value_layer.parameters())
        
        # Create an optimizer with different learning rates for the actor and critic
        self.optimizer = torch.optim.Adam([
            {'params': actor_params, 'lr': policy_lr},
            {'params': critic_params, 'lr': value_lr}
        ], betas=betas)

        self.mse_loss = nn.MSELoss()
        self.wandb_log = wandb_log

    def update(self, states, actions, returns, next_states, dones):
        for _ in range(self.K_epochs):
            # Getting predicted values and log probs for given states and actions
            logprobs, state_values = self.policy.evaluate(states, actions)
            # Calculate the advantages
            advantages = returns - state_values.detach()
            # Normalize the advantages (optional, but can help in training stability)
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-5)
    
            # Compute ratio for PPO
            old_logprobs, _ = self.policy_old.evaluate(states, actions)
            ratio = torch.exp(logprobs - old_logprobs.detach())
    
            # Surrogate loss
            surr1 = ratio * advantages
            surr2 = torch.clamp(ratio, 1-self.eps_clip, 1+self.eps_clip) * advantages
            policy_loss = -torch.min(surr1, surr2).mean()
    
            # Value loss
            value_loss = self.vl_coef * self.mse_loss(state_values, returns)
    
            # Entropy (for exploration)
            entropy_loss = -self.el_coef * logprobs.mean()
    
            # Total loss
            loss = policy_loss + value_loss + entropy_loss
            
            # Optimize policy network
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
            if self.wandb_log:
                wandb.log({
                    "policy_loss": policy_loss.item(),
                    "value_loss": value_loss.item(),
                    "entropy_loss": entropy_loss.item(),
                    "total_loss": loss.item(),
                })
            
        # Copy new weights into old policy
        self.policy_old.load_state_dict(self.policy.state_dict())

    def save(self, path):
        # Saving
        torch.save({
            'model_state_dict': self.policy.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'state_normalizer_state': self.state_normalizer.get_state(),
        }, path)
        
    def load(self, path):
        checkpoint = torch.load(path)
        self.policy.load_state_dict(checkpoint['model_state_dict'])
        self.policy_old.load_state_dict(self.policy.state_dict())
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        self.state_normalizer.set_state(checkpoint['state_normalizer_state'])




import torch
from typing import List

def compute_gae(next_value:float, rewards:List[torch.Tensor],  # List of float32 tensors
                masks:List[torch.Tensor],  # List of float32 tensors
                values:List[torch.Tensor], # List of float32 tensors
                gamma:float, tau:float):
    values = values + [next_value]
    gae = 0
    returns = []
    for step in reversed(range(len(rewards))):
        delta = rewards[step] + gamma * values[step + 1] * masks[step] - values[step]
        gae = delta + gamma * tau * masks[step] * gae
        returns.insert(0, gae + values[step])
    return returns

def compute_returns_and_advantages_without_gae(rewards, states, next_states, dones, value, gamma=0.99):
    """
    Compute returns and advantages without using Generalized Advantage Estimation (GAE).
    
    Parameters:
    - rewards: list of rewards for each timestep
    - states: list of states for each timestep
    - next_states: list of next states for each timestep
    - dones: list of done flags for each timestep
    - value: function to compute the value of a state
    - gamma: discount factor

    Returns:
    - returns: list of computed returns for each timestep
    - advs: list of computed advantages for each timestep
    """
    returns = []
    advs = []

    g = 0
    with torch.no_grad():
        for r, state, next_state, done in zip(
            reversed(rewards),
            reversed(states),
            reversed(next_states),
            reversed(dones),
        ):
            mask = 1 - done.item()
            next_value = value(next_state).item()
            next_value = next_value * mask
            g = r + gamma * next_value
            returns.insert(0, g)
            
            value_curr_state = value(state).item()
            delta = g - value_curr_state
            advs.insert(0, delta)

    return returns, advs

def get_grad_norm(parameters):
    """
    Compute the 2-norm of gradients for the provided parameters.
    
    Args:
    - parameters (Iterable[torch.Tensor]): Network parameters.
    
    Returns:
    - float: Gradient norm.
    """
    grad_norm = 0.0
    for param in parameters:
        if param.grad is not None:
            grad_norm += (param.grad.data ** 2).sum()
    grad_norm = grad_norm.sqrt().item()
    return grad_norm



import numpy as np
import gym
from gym import spaces

class PointMass2DEnv(gym.Env):
    """
    A revised version of the PointMass2DEnv.
    """
    def __init__(self, range=10, max_episode_steps=200, boundary_penalty=5.0):
        super(PointMass2DEnv, self).__init__()
        self.range = range
        self.max_episode_steps = max_episode_steps
        self.boundary_penalty = boundary_penalty
        self.current_step = 0
        self.observation_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.action_space = spaces.Box(low=-1, high=1, shape=(2,), dtype=np.float32)
        self.state = None

    def reset(self):
        self.state = np.array([np.random.uniform(-self.range, self.range), np.random.uniform(-self.range, self.range)], dtype=np.float32)
        self.current_step = 0
        return self._normalize_state(self.state), {}

    def step(self, action):
        self.current_step += 1
        
        # Apply action
        self.state = self.state + action * self.range * 0.1
        reward = 0.5 -np.linalg.norm(self.state) / self.range
        
        # Check boundaries and apply penalty if needed
        if np.any(self.state < -self.range) or np.any(self.state > self.range):
            reward -= self.boundary_penalty
            self.state = np.clip(self.state, -self.range, self.range)

        done = self.current_step >= self.max_episode_steps
        return self._normalize_state(self.state), reward, done, False, {}

    def _normalize_state(self, state):
        return state / self.range

    def render(self, mode='human'):
        pass

    def close(self):
        pass




import torch
import torch.nn as nn
import torch.nn.functional as F


# Define the weight initialization function
def init_weights(m, init_type="default"):
    if isinstance(m, nn.Linear):
        if init_type == "he":
            nn.init.kaiming_normal_(m.weight, nonlinearity="relu")
            nn.init.zeros_(m.bias)
        elif init_type == "xavier":
            nn.init.xavier_normal_(m.weight)
            nn.init.zeros_(m.bias)
        elif init_type == "default":
            # The default initialization in PyTorch for Linear layers
            # is almost like Xavier uniform. We'll simply skip any
            # custom initialization to rely on the default.
            pass
        else:
            raise ValueError(f"Initialization type {init_type} not recognized.")

class PolicyNetwork(nn.Module):
    def __init__(self, state_dim, action_dim, n_latent_var, init_type="he"):
        super(PolicyNetwork, self).__init__()

        # Actor: outputs mean and log standard deviation
        self.action_mu = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        self.action_log_std = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        # Apply the initialization
        self.apply(lambda m: init_weights(m, init_type=init_type))

    def forward(self, x):
        mean = self.action_mu(x)
        log_std = self.action_log_std(x)
        
        # Compute std using the exponential of log_std
        std = torch.exp(log_std)
        dist = torch.distributions.normal.Normal(mean, std)
        return dist

class ValueNetwork(nn.Module):
    def __init__(self, state_size, hidden_size, init_type):
        super(ValueNetwork, self).__init__()
        self.fc1 = nn.Linear(state_size, hidden_size)
        self.fc2 = nn.Linear(hidden_size, hidden_size)
        self.fc3 = nn.Linear(hidden_size, 1)

        # Apply the initialization
        self.apply(lambda m: init_weights(m, init_type=init_type))

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        x = self.fc3(x)
        return x



import gym
from .envs.point_mass1d import PointMass1DEnv
from .envs.point_mass2d import PointMass2DEnv

gym.register("PointMass1D-v0", entry_point=PointMass1DEnv, max_episode_steps=200)
gym.register("PointMass2D-v0", entry_point=PointMass2DEnv, max_episode_steps=200)


import os
import glob
import torch


class CheckpointManager:
    @staticmethod
    def save_checkpoint(policy, value, optimizer, normalizer, epoch, checkpoint_path):
        # Create checkpoint directory if it does not exist
        if not os.path.exists(checkpoint_path):
            os.makedirs(checkpoint_path)
        checkpoint = {
            "epoch": epoch,
            "policy_state_dict": policy.state_dict(),
            "value_state_dict": value.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "normalizer_state": normalizer.get_state(),
        }
        torch.save(checkpoint, f"{checkpoint_path}/checkpoint_epoch{epoch}.pt")

    @staticmethod
    def load_checkpoint(policy, value, optimizer, normalizer, checkpoint_path, epoch=None):
        # Find the latest checkpoint file
        if epoch is None:
            checkpoint_files = sorted(
                glob.glob(f"{checkpoint_path}/checkpoint_epoch*.pt")
            )
            if len(checkpoint_files) == 0:
                raise ValueError("No checkpoint found in the specified directory.")
            checkpoint_file = checkpoint_files[-1]
        else:
            checkpoint_file = f"{checkpoint_path}/checkpoint_epoch{epoch}.pt"

        checkpoint = torch.load(checkpoint_file)
        policy.load_state_dict(checkpoint["policy_state_dict"])
        value.load_state_dict(checkpoint["value_state_dict"])
        optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
        normalizer.set_state(checkpoint["normalizer_state"])
        epoch = checkpoint["epoch"]
        return epoch



import torch
from torch import optim
import torch.nn as nn
from torch.nn import functional as F
from time import time
from torch.optim.lr_scheduler import ExponentialLR, CosineAnnealingLR
import numpy as np
from tqdm import tqdm
import gym
import os
from nanoppo.environment_manager import EnvironmentManager
from nanoppo.network_manager import NetworkManager
from nanoppo.checkpoint_manager import CheckpointManager
from nanoppo.wandb_logger import WandBLogger
from nanoppo.policy.network import PolicyNetwork, ValueNetwork
from nanoppo.rollout_buffer import RolloutBuffer
from nanoppo.reward_scaler import RewardScaler
from nanoppo.reward_shaper import RewardShaper, TDRewardShaper
from nanoppo.normalizer import Normalizer
from nanoppo.state_scaler import StateScaler
from nanoppo.metrics_recorder import MetricsRecorder
from nanoppo.ppo_utils import compute_gae, compute_returns_and_advantages_without_gae, get_grad_norm

import warnings

# Suppress the specific warning
warnings.filterwarnings(
    "ignore",
    message="Could not parse CUBLAS_WORKSPACE_CONFIG, using default workspace size of 8519680 bytes.",
)


class PPOAgent:
    def __init__(self, config, optimizer_config, force_cpu=False):
        self.config = config
        self.optimizer_config = optimizer_config
        if force_cpu:
            self.device = torch.device("cpu")
        else:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.project = config["project"]
        self.env_name = config["env_name"]
        self.env_manager = EnvironmentManager(self.env_name, config["env_config"])
        self.env = self.env_manager.setup_env()

        self.network_manager = NetworkManager(
            self.env,
            optimizer_config,
            config["hidden_size"],
            config["init_type"],
            self.device,
        )
        (
            self.policy,
            self.value,
            self.optimizer,
            self.scheduler,
            self.policy_old,
            self.value_old
        ) = self.network_manager.setup_networks()

        self.rollout_buffer = RolloutBuffer(self.config["batch_size"])
        if self.config["rescaling_rewards"]:
            self.reward_scaler = RewardScaler()
        else:
            self.reward_scaler = None
        
        self.normalizer = None
        self.state_scaler = None
        if self.config["scale_states"] == 'default':
            self.normalizer = Normalizer(self.env.observation_space.shape[0])
        elif self.config["scale_states"] in ["env", "standard", "minmax", "robust", "quantile"]:
            self.state_scaler = StateScaler(
                self.env, sample_size=10000, scale_type=self.config["scale_states"]
            )
        else:
            raise ValueError(f"Unknown scale type: {self.config['scale_states']}")
        
        self.metrics_log = self.config["metrics_log"]
        if self.metrics_log:
            self.metrics_recorder = MetricsRecorder()
        else:
            self.metrics_recorder = None

        self.project = self.config["project"]
        self.wandb_log = self.config["wandb_log"]
        # Initialize WandB
        if self.wandb_log:
            config = locals().copy()
            keys_to_delete = [
                "checkpoint_interval",
                "checkpoint_path",
                "resume_training",
                "resume_epoch",
                "report_func",
            ]
            [config.pop(key, None) for key in keys_to_delete]
            WandBLogger.init(project=self.project, name=self.env_name, config=config)
        self.checkpoint_interval = self.config["checkpoint_interval"]
        self.checkpoint_dir = os.path.join(
            self.config["checkpoint_dir"], self.project, self.env_name
        )
        self.log_interval = self.config["log_interval"]

    @staticmethod
    def get_epoch_iterator(last_epoch, epochs, verbose: int):
        if verbose > 0:
            progress_iterator = tqdm(range(last_epoch, last_epoch + epochs))
        else:
            progress_iterator = range(last_epoch, last_epoch + epochs)
        return progress_iterator

    @staticmethod
    def load_checkpoint(self, checkpoint_path, epoch=None):
        last_epoch = CheckpointManager.load_checkpoint(
            self.policy, self.value, self.optimizer, self.normalizer, checkpoint_path, epoch
        )
        return last_epoch

    @staticmethod
    def select_action(policy: PolicyNetwork, state, device, action_min, action_max):
        state = torch.FloatTensor(state).to(device)
        dist = policy(state)
        action = dist.sample()
        # If action space is continuous, compute the log_prob of the action, sum(-1) to sum over all dimensions
        log_prob = dist.log_prob(action).sum(-1)

        # Extract mean and std of the action distribution
        action_mean = dist.mean
        action_std = dist.stddev

        # action = torch.tanh(action)  # Pass the sampled action through the tanh activation function
        # action = action + torch.normal(mean=torch.zeros_like(action), std=action_std)  # Add noise to the action
        #action = action.clamp(
        #    action_min, action_max
        #)  # Clip the action to the valid range of the action space
        return (
            action.cpu().detach().numpy(),
            log_prob.cpu().detach().numpy(),
            action_mean.cpu().detach().numpy(),
            action_std.cpu().detach().numpy(),
        )

    @staticmethod
    def surrogate(policy, old_probs, states, actions, advs, clip_param, entropy_coef):
        # Policy loss
        dist = policy(states)
        new_probs = dist.log_prob(actions).sum(-1)
        ratio = torch.exp(new_probs - old_probs)  # Importance sampling ratio
        surr1 = ratio * advs
        surr2 = torch.clamp(ratio, 1 - clip_param, 1 + clip_param) * advs # Trust region clipping
        #entropy = dist.entropy().mean()  # Compute the mean entropy of the distribution
        #entropy_loss = - entropy_coef * entropy

        # Entropy (for exploration)
        approximate_entropy_loss = - entropy_coef * new_probs.mean()

        return (
            -torch.min(surr1, surr2).mean(),
            approximate_entropy_loss
        )  # Add the entropy term to the policy loss

    @staticmethod
    def compute_value_loss(state_values, returns):
        # Compute value loss
        v_pred = state_values
        v_target = returns.squeeze()
        value_loss =  F.mse_loss(v_pred, v_target)
        return value_loss

    @staticmethod
    def minibatch_update(
        iter_num,
        policy,
        value,
        policy_old,
        value_old,
        optimizer,
        scheduler,
        rollout_buffer: RolloutBuffer,
        device,
        batch_size,
        sgd_iters,
        gamma,
        clip_param,
        vf_coef,
        entropy_coef,
        max_grad_norm,
        use_gae,
        tau,
        wandb_log,
        metrics_recorder: MetricsRecorder,
    ):
        (
            batch_states,
            batch_actions,
            batch_log_probs,
            batch_rewards,
            batch_next_states,
            batch_dones,
        ) = rollout_buffer.sample(batch_size, device=device, randomize=False)

        # Compute returns once from rollout buffer in order 
        if use_gae:
            # Compute Advantage using GAE and Returns
            values = value(batch_states).detach().squeeze().tolist() # For values + [next_value]
            next_value = value(batch_next_states[-1]).item()
            masks = [1 - done.item() for done in batch_dones]
            # Only compute returns once
            returns = compute_gae(
                next_value, batch_rewards, masks, values, gamma, tau
            )
        else:
            returns, advs = compute_returns_and_advantages_without_gae(
                batch_rewards,
                batch_states,
                batch_next_states,
                batch_dones,
                value,
                gamma,
            )

        returns = torch.tensor(returns, dtype=torch.float32).to(device)
        for sgd_iter in range(sgd_iters):
            # Compute advantages separately for each SGD iteration
            state_values = value(batch_states).squeeze()
            advantages = returns - state_values.detach()
            # Normalize the advantages (optional, but can help in training stability)
            advantages = (advantages - advantages.mean()) / (advantages.std() + 1e-5)

            policy_loss, entropy_loss = PPOAgent.surrogate(
                policy,
                old_probs=batch_log_probs,
                states=batch_states,
                actions=batch_actions,
                advs=advantages,
                clip_param=clip_param,
                entropy_coef=entropy_coef,
            )

            value_loss = PPOAgent.compute_value_loss(state_values, returns)

            # Compute total loss and update parameters
            total_loss = policy_loss + entropy_loss + vf_coef * value_loss

            optimizer.zero_grad()
            total_loss.backward()
            optimizer.step()
            #if scheduler is not None:
            #    scheduler.step()

            """
            # Clip the gradients to avoid exploding gradients
            policy_grad_norm = nn.utils.clip_grad_norm_(
                policy.parameters(), max_grad_norm
            )
            value_grad_norm = nn.utils.clip_grad_norm_(
                value.parameters(), max_grad_norm
            )
            """
            action_mu_grad_norm = get_grad_norm(policy.action_mu.parameters())
            action_log_std_grad_norm = get_grad_norm(policy.action_log_std.parameters())
            value_grad_norm = get_grad_norm(value.parameters())

            # compute activation norm
            # remove the forward hooks
            """
            for hook in hooks:
                hook.remove()
            # compute the mean activation norm for the value network
            activation_norm = sum(activation_norms) / len(activation_norms)
            """


            if wandb_log:
                # Log the losses and gradients to WandB
                WandBLogger.log(
                    {
                        "iteration": iter_num,
                        "Loss/Total": total_loss.item(),
                        "Loss/Policy": policy_loss.item(),
                        "Loss/Entropy": entropy_loss.item(),
                        "Loss/Value": value_loss.item(),
                        "Loss/Coef_Value": vf_coef * value_loss.item(),
                    }
                )
                # wandb.log({"Gradients/PolicyNet": wandb.Histogram(policy.fc1.weight.grad.detach().cpu().numpy())})
                WandBLogger.log(
                    {
                        "Policy/Mu_Gradient_Norm": action_mu_grad_norm,
                        "Policy/Log_Std_Gradient_Norm": action_log_std_grad_norm,
                        "Value/Gradient_Norm": value_grad_norm,
                        # "Value/Activation_Norm": activation_norm
                    }
                )
                # wandb.log({"Gradients/ValueNet": wandb.Histogram(value.fc1.weight.grad.detach().cpu().numpy())})
                log_std_value = policy.action_log_std(batch_states).detach().cpu().numpy()
                WandBLogger.log({"Policy/Log_Std": log_std_value})
            # log the learning rate to wandb
            lrs = {}
            for i, param_group in enumerate(optimizer.param_groups):
                lr = param_group["lr"]
                lrs["learning_rate_{}".format(i)] = lr
                if wandb_log:
                    WandBLogger.log({"LR/LearningRate_{}".format(i): lr})

            if metrics_recorder:
                metrics_recorder.record_losses(
                    total_loss.item(),
                    policy_loss.item(),
                    entropy_loss.item(),
                    value_loss.item(),
                )
                metrics_recorder.record_learning(lrs)

            iter_num += 1

        rollout_buffer.clear()  # clear the rollout buffer, all data is from the current policy
        assert(len(rollout_buffer) == 0)
        # Copy new weights into old policy
        policy_old.load_state_dict(policy.state_dict())
        return (
            policy,
            value,
            iter_num,
        )

    @staticmethod
    def train_with_epoch(
        project:str,
        env: gym.Env,
        env_name: str,
        env_config: dict,
        rollout_buffer: RolloutBuffer,
        state_scaler: StateScaler,
        shape_reward: RewardShaper,
        normalizer: Normalizer,
        reward_scaler: RewardScaler,
        policy: PolicyNetwork,
        value: ValueNetwork,
        policy_old,
        value_old,
        optimizer: optim.Optimizer,
        scheduler: optim.lr_scheduler._LRScheduler,
        epochs: int,
        max_timesteps: int,
        batch_size: int,
        sgd_iters: int,
        gamma: float,
        clip_param: float,
        vf_coef: float,
        entropy_coef: float,
        max_grad_norm: float,
        use_gae: bool,
        tau: float,
        verbose: int,
        checkpoint_interval:int,
        checkpoint_dir: str,
        log_interval: int,
        resume_training: bool,
        resume_epoch: int,
        report_func: callable = None,
        device: str = "cpu",
        wandb_log: bool = True,
        metrics_recorder: MetricsRecorder = None,
    ):
        checkpoint_path = os.path.join(checkpoint_dir, project, env_name)
        if resume_training:
            if verbose > 0:
                print("Resuming training from checkpoint...", checkpoint_path, 'from latest' if resume_epoch <= 0 else f'from epoch {resume_epoch}')
            last_epoch = CheckpointManager.load_checkpoint(
                policy, value, optimizer, checkpoint_path, None if resume_epoch <=0 else resume_epoch - 1
            )
            last_epoch += 1
        else:
            last_epoch = 0

        if shape_reward is None:
            reward_shaper = None
        elif shape_reward == TDRewardShaper:
            reward_shaper = TDRewardShaper(model=value, device=device)
        else:
            reward_shaper = shape_reward()

        # Set up training loop
        total_iters = epochs * sgd_iters
        time_steps = 0
        episode_steps = []
        episode_rewards = []  # Initialize cumulative reward
        train_iters = 0
        average_reward = -np.inf
        best_reward = -np.inf
        start = time()
        for epoch in PPOAgent.get_epoch_iterator(last_epoch, epochs, verbose):
            state, info = env.reset()
            if isinstance(state, dict):
                state = state["obs"]
            if normalizer:
                normalizer.observe(state)
                scaled_state = normalizer.normalize(state)
            elif state_scaler:
                scaled_state = state_scaler.scale_state(state)
            else:
                raise ValueError("No state scaler or normalizer is provided")
            done = False
            total_reward = 0
            
            for step in range(max_timesteps):
                action, log_prob, action_mean, action_std = PPOAgent.select_action(
                    policy,
                    scaled_state,
                    device,
                    env.action_space.low[0],
                    env.action_space.high[0],
                )
                if wandb_log:
                    WandBLogger.log_action_distribution_parameters(
                        action_mean, action_std
                    )

                next_state, reward, done, truncated, info = env.step(action)

                if isinstance(next_state, dict):
                    next_state = next_state["obs"]
                if normalizer:
                    normalizer.observe(next_state)
                    scaled_next_state = normalizer.normalize(next_state)
                elif state_scaler:
                    scaled_next_state = state_scaler.scale_state(next_state)
                else:
                    raise ValueError("No state scaler or normalizer is provided")

                scaled_state = scaled_next_state
                total_reward += reward
                # Reshape rewards
                if reward_shaper is None:
                    reshaped_reward = reward
                else:
                    raise NotImplementedError
                    reshaped_reward = reward_shaper.reshape(
                        [reward], [state], [next_state]
                    )[0]
                # Scale rewards
                if reward_scaler is None:
                    scaled_reward = reshaped_reward
                    # click.secho("Warning: Reward scaling is not applied.", fg="yellow", err=True)
                else:
                    raise NotImplementedError
                    scaled_reward = reward_scaler.scale_rewards([reshaped_reward])[0]
                rollout_buffer.push(
                    state=scaled_state,
                    action=action.squeeze(),
                    log_prob=log_prob,
                    reward=scaled_reward,
                    next_state=scaled_next_state,
                    done=done,
                )
                time_steps += 1
                if time_steps % batch_size == 0:
                    _, _, train_iters = PPOAgent.minibatch_update(
                        train_iters,
                        policy,
                        value,
                        policy_old,
                        value_old,
                        optimizer,
                        scheduler,
                        rollout_buffer,
                        device,
                        batch_size,
                        sgd_iters,
                        gamma,
                        clip_param,
                        vf_coef,
                        entropy_coef,
                        max_grad_norm=max_grad_norm,
                        use_gae=use_gae,
                        tau=tau,
                        wandb_log=wandb_log,
                        metrics_recorder=metrics_recorder,
                    )
 
                if done or truncated:
                    break

            episode_steps.append(step + 1)
            episode_rewards.append(total_reward) 
                
            # Average of last 10 episodes or all episodes if less than 10 episodes are available
            if len(episode_rewards) >= 20 and (epoch + 1) % log_interval == 0:
                average_reward = sum(episode_rewards[-20:]) / len(episode_rewards[-20:])
                if wandb_log:
                    WandBLogger.log_rewards(episode_rewards[-20:])
                if metrics_recorder:
                    metrics_recorder.record_rewards(episode_rewards[-20:])
                if report_func:
                    report_func(mean_reward=average_reward)  # Reporting the reward

            if verbose > 0 and (epoch + 1) % log_interval == 0:
                print("env", env_name,
                    "epoch",
                    epoch + 1,
                    "reward episodes",
                    len(episode_rewards),
                    "everage steps",
                    np.mean(episode_steps),
                    "average reward",
                    round(average_reward,2),
                    "train epochs",
                    epoch - last_epoch + 1,
                    "train iters",
                    train_iters,
                    "rollout episodes",
                    len(episode_rewards),
                    "rollout time steps",
                    time_steps,
                )

                action_mu_grad_norm = get_grad_norm(policy.action_mu.parameters())
                action_log_std_grad_norm = get_grad_norm(policy.action_log_std.parameters())
                value_grad_norm = get_grad_norm(value.parameters())
                print('action_mu_grad_norm', round(action_mu_grad_norm,2), 'action_log_std_grad_norm', round(action_log_std_grad_norm,2), 
                      'value_grad_norm', round(value_grad_norm,2))

            if checkpoint_interval > 0: 
                if average_reward > best_reward:
                    print("Saving checkpoint...", checkpoint_path)
                    print('avg_reward', average_reward, '> best_reward', best_reward)
                    best_reward = average_reward
                    CheckpointManager.save_checkpoint(
                    policy, value, optimizer, normalizer, epoch, checkpoint_path)

        end = time()
        print("Training time: ", round((end - start) / 60, 2), "minutes")
        if metrics_recorder:
            metrics_recorder.to_csv()
        if wandb_log:
            WandBLogger.finish()
        print(
            "Training complete",
            "average reward",
            average_reward,
            "total iters",
            total_iters,
        )
        return policy, value, average_reward, train_iters

    def train(self, epochs):
        PPOAgent.train_with_epoch(
            project=self.project,
            env=self.env,
            env_name=self.env_name,
            env_config=self.config["env_config"],
            rollout_buffer=self.rollout_buffer,
            state_scaler=self.state_scaler,
            shape_reward=self.config["shape_reward"],
            reward_scaler=self.reward_scaler,
            normalizer=self.normalizer,
            policy=self.policy,
            value=self.value,
            policy_old = self.policy_old,
            value_old = self.value_old,
            optimizer=self.optimizer,
            scheduler=self.scheduler,
            epochs=epochs,
            max_timesteps=self.config["max_timesteps"],
            batch_size=self.config["batch_size"],
            sgd_iters=self.config["sgd_iters"],
            gamma=self.config["gamma"],
            # optimizer_config=self.optimizer_config,
            # hidden_size=self.config["hidden_size"],
            # init_type=self.config["init_type"],
            clip_param=self.config.get("clip_param", 0.2),
            vf_coef=self.config["vf_coef"],
            entropy_coef=self.config["entropy_coef"],
            max_grad_norm=self.config["max_grad_norm"],
            use_gae=self.config["use_gae"],
            tau=self.config["tau"],
            verbose=self.config["verbose"],
            checkpoint_interval=self.config["checkpoint_interval"],
            checkpoint_dir=self.config['checkpoint_dir'],
            log_interval=self.config["log_interval"],
            resume_training=self.config["resume_training"],
            resume_epoch=self.config["resume_epoch"],
            report_func=self.config["report_func"],
            # project=self.config["project"],
            device=self.device,
            wandb_log=self.config["wandb_log"],
            metrics_recorder=self.metrics_recorder,
            # seed=self.config["seed"],
        )



import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import numpy as np


class PolicyNetwork(nn.Module):
    def __init__(self, input_dim, hidden_dim, n_actions):
        super(PolicyNetwork, self).__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, n_actions)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        return F.softmax(self.fc2(x), dim=-1)


class ValueNetwork(nn.Module):
    def __init__(self, input_dim, hidden_dim):
        super(ValueNetwork, self).__init__()
        self.fc1 = nn.Linear(input_dim, hidden_dim)
        self.fc2 = nn.Linear(hidden_dim, 1)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        return self.fc2(x)


class PPO:
    def __init__(
        self, state_dim, action_dim, hidden_dim=64, lr=0.001, gamma=0.99, epsilon=0.2
    ):
        self.policy = PolicyNetwork(state_dim, hidden_dim, action_dim)
        self.value = ValueNetwork(state_dim, hidden_dim)
        self.policy_optimizer = optim.Adam(self.policy.parameters(), lr=lr)
        self.value_optimizer = optim.Adam(self.value.parameters(), lr=lr)
        self.gamma = gamma
        self.epsilon = epsilon

    def compute_advantage(self, rewards, values, next_value, done):
        advantages = torch.zeros_like(rewards)
        returns = next_value
        for t in reversed(range(len(rewards))):
            returns = rewards[t] + self.gamma * returns * (1 - done[t])
            advantages[t] = returns - values[t]
        return advantages

    def update(self, states, actions, rewards, next_states, done, log_probs):
        states = torch.tensor(np.array(states), dtype=torch.float32)
        actions = torch.tensor(actions, dtype=torch.long)
        rewards = torch.tensor(rewards, dtype=torch.float32)
        next_states = torch.tensor(np.array(next_states), dtype=torch.float32)
        done = torch.tensor(done, dtype=torch.float32)
        old_log_probs = torch.tensor(log_probs, dtype=torch.float32)

        # Compute advantage
        values = self.value(states).squeeze()
        if done[-1]:
            next_value = 0
        else:
            next_value = self.value(next_states[-1]).item()
        advantages = self.compute_advantage(
            rewards, values, next_value, done
        ).detach()  # Detach advantages

        # Compute value loss and update value network
        targets = rewards + self.gamma * self.value(next_states).squeeze() * (1 - done)
        value_loss = F.mse_loss(values, targets.detach())
        self.value_optimizer.zero_grad()
        value_loss.backward()
        self.value_optimizer.step()

        # Compute new log probabilities
        new_log_probs = (
            self.policy(states).gather(1, actions.unsqueeze(1)).log().squeeze()
        )

        # Compute probability ratio
        ratio = torch.exp(new_log_probs - old_log_probs)

        # Compute surrogate loss
        surr1 = ratio * advantages
        surr2 = torch.clamp(ratio, 1 - self.epsilon, 1 + self.epsilon) * advantages
        policy_loss = -torch.min(surr1, surr2).mean()

        # Update policy network
        self.policy_optimizer.zero_grad()
        policy_loss.backward()
        self.policy_optimizer.step()



import torch
from torch import optim
import gym
from nanoppo.policy.network import PolicyNetwork, ValueNetwork
from nanoppo.policy.actor_critic import ActorCritic
from torch.optim.lr_scheduler import ExponentialLR, CosineAnnealingLR


class NetworkManager:
    def __init__(self, env, optimizer_config, hidden_size, init_type, device, network_type="actor_critic"):
        self.network_type = network_type
        self.env = env
        self.optimizer_config = optimizer_config
        self.hidden_size = hidden_size
        self.init_type = init_type
        self.device = device

    def setup_networks(self):
        if isinstance(self.env.observation_space, gym.spaces.Dict):
            observation_space = self.env.observation_space.spaces["obs"]
        else:
            observation_space = self.env.observation_space
        if isinstance(self.env.action_space, gym.spaces.Box):
            action_dim = self.env.action_space.shape[0]
        else:
            action_dim = self.env.action_space.n

        if self.network_type == "actor_critic":
            policy = ActorCritic(
                state_dim=observation_space.shape[0],
                action_dim=action_dim,
                n_latent_var=self.hidden_size,
                action_low=self.env.action_space.low,
                action_high=self.env.action_space.high
            ).float().to(self.device)

            value = policy.value_layer
            
            # Separate the parameters of the actor and critic networks
            actor_params = list(policy.action_mu.parameters()) + list(policy.action_log_std.parameters())
            critic_params = list(policy.value_layer.parameters())

            policy_old = ActorCritic(
                state_dim=observation_space.shape[0],
                action_dim=action_dim,
                n_latent_var=self.hidden_size,
                action_low=self.env.action_space.low,
                action_high=self.env.action_space.high
            ).float().to(self.device)
            policy_old.load_state_dict(policy.state_dict())
            value_old = policy_old.value_layer
        
        else:
            policy = PolicyNetwork(
                observation_space.shape[0], action_dim, init_type=self.init_type
            ).to(self.device)
            value = ValueNetwork(
                observation_space.shape[0],
                hidden_size=self.hidden_size,
                init_type=self.init_type,
            ).to(self.device)
            actor_params = list(policy.parameters())
            critic_params = list(value.parameters())
    
        policy_lr = self.optimizer_config["policy_lr"]
        value_lr = self.optimizer_config["value_lr"]
    
        optimizer = optim.Adam(
            [
                {"params": actor_params, "lr": policy_lr},
                {"params": critic_params, "lr": value_lr},
            ],
            betas=(self.optimizer_config["beta1"], self.optimizer_config["beta2"]),
            eps=self.optimizer_config["epsilon"],
            weight_decay=self.optimizer_config["weight_decay"],
        )
    
        if self.optimizer_config["scheduler"] is None:
            scheduler = None
        elif self.optimizer_config["scheduler"] == "exponential":
            scheduler = ExponentialLR(
                optimizer, gamma=self.optimizer_config["exponential_gamma"]
            )
        elif self.optimizer_config["scheduler"] == "cosine":
            scheduler = CosineAnnealingLR(
                optimizer, T_max=self.optimizer_config["cosine_T_max"]
            )
        else:
            raise ValueError(
                f"Scheduler {self.optimizer_config['scheduler']} not recognized."
            )

        return policy, value, optimizer, scheduler, policy_old, value_old



class RewardShaper:
    """
    Base class for reward shaping. It provides a structure and can be subclassed
    for specific environments.
    """

    def __init__(self):
        pass

    def reshape(self, rewards, observations, next_observations=None):
        """
        This method should be overridden by subclasses to provide specific reward shaping logic.

        Parameters:
        - rewards (list or array): Original rewards from the environment.
        - observations (list or array): Corresponding observations for the rewards.

        Returns:
        - Reshaped rewards.
        """
        # By default, just return the original rewards.
        return rewards


class MountainCarRewardShaper(RewardShaper):
    """
    Reward shaping specifically designed for the MountainCar environment.
    """

    def __init__(self, position_weight=1.0, velocity_weight=1.0):
        """
        Initialize the reward shaper with given weights.

        Parameters:
        - position_weight (float): Weight for position-based reward shaping.
        - velocity_weight (float): Weight for velocity-based reward shaping.
        """
        super().__init__()
        self.position_weight = position_weight
        self.velocity_weight = velocity_weight

    def reshape(self, rewards, observations, next_observations=None):
        """
        Reward shaping for MountainCar based on position and velocity.

        Parameters:
        - rewards (list or array): Original rewards from the environment.
        - observations (list or array): Corresponding observations for the rewards.

        Returns:
        - Reshaped rewards.
        """
        # Extract position and velocity from observations.
        positions = [obs[0] for obs in observations]
        velocities = [obs[1] for obs in observations]

        # Calculate position-based and velocity-based rewards.
        position_rewards = [pos * self.position_weight for pos in positions]
        velocity_rewards = [abs(vel) * self.velocity_weight for vel in velocities]

        # Combine the original rewards with the shaped rewards.
        reshaped_rewards = [
            orig_reward + pos_reward + vel_reward
            for orig_reward, pos_reward, vel_reward in zip(
                rewards, position_rewards, velocity_rewards
            )
        ]

        return reshaped_rewards


import numpy as np


class MountainCarHeightRewardShaper(RewardShaper):
    """
    Reward shaping specifically designed for the MountainCar environment based on the car's height.
    """

    def __init__(self, height_weight=1.0):
        """
        Initialize the reward shaper with given weight.

        Parameters:
        - height_weight (float): Weight for height-based reward shaping.
        """
        super().__init__()
        self.height_weight = height_weight

    def height(self, position):
        """
        Calculate the height of the car based on its position.

        Parameters:
        - position (float): Current position of the car.

        Returns:
        - Height of the car.
        """
        # This formula is derived from the MountainCar environment's terrain shape.
        return np.sin(3 * position) * 0.45 + 0.55

    def reshape(self, rewards, observations, next_observations=None):
        """
        Reward shaping for MountainCar based on the car's height.

        Parameters:
        - rewards (list or array): Original rewards from the environment.
        - observations (list or array): Corresponding observations for the rewards.

        Returns:
        - Reshaped rewards.
        """
        # Calculate height-based rewards.
        height_rewards = [
            self.height(obs[0]) * self.height_weight for obs in observations
        ]

        # Combine the original rewards with the shaped rewards.
        reshaped_rewards = [
            orig_reward + height_reward
            for orig_reward, height_reward in zip(rewards, height_rewards)
        ]

        return reshaped_rewards


class MountainCarAdvancedRewardShaper(RewardShaper):
    """
    Advanced reward shaping for the MountainCar environment, combining multiple techniques.
    """

    def __init__(
        self,
        height_weight=1.0,
        velocity_weight=1.0,
        still_penalty=-0.5,
        still_threshold=0.01,
    ):
        """
        Initialize the reward shaper with given weights and parameters.

        Parameters:
        - height_weight (float): Weight for height-based reward shaping.
        - velocity_weight (float): Weight for velocity-based reward shaping.
        - still_penalty (float): Penalty for the car when it's nearly still.
        - still_threshold (float): Velocity threshold to consider the car as being 'still'.
        """
        super().__init__()
        self.height_weight = height_weight
        self.velocity_weight = velocity_weight
        self.still_penalty = still_penalty
        self.still_threshold = still_threshold

    def height(self, position):
        """
        Calculate the height of the car based on its position.

        Parameters:
        - position (float): Current position of the car.

        Returns:
        - Height of the car.
        """
        return np.sin(3 * position) * 0.45 + 0.55

    def reshape(self, rewards, observations, next_observations=None):
        """
        Advanced reward shaping for MountainCar based on multiple techniques.

        Parameters:
        - rewards (list or array): Original rewards from the environment.
        - observations (list or array): Corresponding observations for the rewards.

        Returns:
        - Reshaped rewards.
        """
        reshaped_rewards = []
        for reward, obs in zip(rewards, observations):
            position, velocity = obs

            # Height-based reward
            height_reward = self.height(position) * self.height_weight

            # Velocity-based reward
            velocity_reward = abs(velocity) * self.velocity_weight

            # Penalty for staying still
            penalty = self.still_penalty if abs(velocity) < self.still_threshold else 0

            # Combine the rewards
            total_reward = reward + height_reward + velocity_reward + penalty
            reshaped_rewards.append(total_reward)

        return reshaped_rewards


class MountainCarDirectionalRewardShaper(RewardShaper):
    """
    Reward shaping for MountainCar based on height, velocity direction, and distance from the goal.
    """

    def __init__(
        self, height_weight=1.0, velocity_direction_weight=1.0, distance_weight=1.0
    ):
        super().__init__()
        self.height_weight = height_weight
        self.velocity_direction_weight = velocity_direction_weight
        self.distance_weight = distance_weight

    def height(self, position):
        return np.sin(3 * position) * 0.45 + 0.55

    def reshape(self, rewards, observations, next_observations=None):
        reshaped_rewards = []

        for reward, obs in zip(rewards, observations):
            position, velocity = obs

            # Height-based reward
            height_reward = self.height(position) * self.height_weight

            # Velocity direction reward
            if (position < 0 and velocity > 0) or (position > 0 and velocity < 0):
                direction_reward = abs(velocity) * self.velocity_direction_weight
            else:
                direction_reward = 0

            # Distance from the goal reward
            goal_position = 0.5
            distance_reward = -abs(goal_position - position) * self.distance_weight

            # Combine the rewards
            total_reward = reward + height_reward + direction_reward + distance_reward
            reshaped_rewards.append(total_reward)

        return reshaped_rewards


import torch


class TDRewardShaper(RewardShaper):
    """
    Reward shaping for MountainCar based on Temporal Difference (TD) error.
    """

    def __init__(self, model, device, gamma=0.99):
        super().__init__()
        self.model = model  # The neural network model used by PPO to estimate values
        self.device = device
        self.gamma = gamma

    def td_error(self, reward, current_state, next_state):
        current_state = torch.from_numpy(current_state).float().to(self.device)
        current_value = self.model(current_state)
        next_state = torch.from_numpy(next_state).float().to(self.device)
        next_value = self.model(next_state)

        return reward + self.gamma * next_value - current_value

    def reshape(self, rewards, observations, next_observations):
        reshaped_rewards = []
        with torch.no_grad():
            for t in range(len(rewards)):
                reward = rewards[t]
                current_state = observations[t]
                next_state = next_observations[t]
                td = self.td_error(reward, current_state, next_state)
                r = (reward + td).cpu().numpy().item()
                reshaped_rewards.append(r)
        return reshaped_rewards



import torch
import torch.nn as nn

class ActorCritic(nn.Module):
    def __init__(self, state_dim, action_dim, n_latent_var, action_low, action_high):
        super(ActorCritic, self).__init__()
        self.action_low = action_low
        self.action_high = action_high

        # Actor: outputs mean and log standard deviation
        self.action_mu = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        self.action_log_std = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, action_dim)
        )

        # Critic
        self.value_layer = nn.Sequential(
            nn.Linear(state_dim, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, n_latent_var),
            nn.Tanh(),
            nn.Linear(n_latent_var, 1)
        )

    def forward(self, state):
        mu = self.action_mu(state)
        log_std = self.action_log_std(state)
        std = log_std.exp()
        dist = torch.distributions.Normal(mu, std)
        return dist

    def act(self, state, action=None, compute_logprobs=True):
        mu = self.action_mu(state)
        log_std = self.action_log_std(state)
        std = log_std.exp()
        dist = torch.distributions.Normal(mu, std)
        if action is None:
            action = dist.sample()
            # Assuming env.action_space.low and env.action_space.high are numpy arrays
            action_low_tensor = torch.tensor(self.action_low, dtype=torch.float32)
            action_high_tensor = torch.tensor(self.action_high, dtype=torch.float32)
            #action = torch.clamp(action, action_low_tensor, action_high_tensor)

        if compute_logprobs:
            logprobs = dist.log_prob(action).sum(axis=-1)
            return action, logprobs
        
        return action

    def evaluate(self, state, action):
        mu = self.action_mu(state)
        log_std = self.action_log_std(state)
        std = log_std.exp()
        
        dist = torch.distributions.Normal(mu, std)
        logprobs = dist.log_prob(action).sum(axis=-1)
        state_value = self.value_layer(state)
        
        return logprobs, torch.squeeze(state_value)

    def get_value(self, state):
        return self.value_layer(state)



import gym
from gym import spaces
import numpy as np

class PointMass1DEnv(gym.Env):
    def __init__(self, action_range=1.0, position_limit=10.0, damping_factor=0.9, max_episode_steps=200):
        super(PointMass1DEnv, self).__init__()
        self.action_range = action_range
        # Action space: Force [-1.0, 1.0]
        self.action_space = spaces.Box(low=-action_range, high=action_range, shape=(1,), dtype=float)
        
        # State space: Position and Velocity
        self.observation_space = spaces.Box(low=-float(position_limit), high=float(position_limit), shape=(2,), dtype=float)
        
        # Parameters
        self.max_steps = max_episode_steps
        self.current_step = 0
        self.damping_factor = damping_factor

    def reset(self):
        self.state = np.array([0.5 * (2 * np.random.rand() - 1), 0.0], dtype=np.float32)  # random initial position, zero velocity
        self.current_step = 0
        return self.state, {}

    def step(self, action):
        position, velocity = self.state

        # Update velocity and position based on the force (action)
        # Update velocity and position based on the force (action)
        velocity = self.damping_factor * (velocity + action[0])  # Introducing damping
        position += velocity
        
        # Clamp position to bounds
        position = np.clip(position, self.observation_space.low[0], self.observation_space.high[0])

        self.state = np.array([position, velocity], dtype=np.float32)

        # Calculate reward
        reward = 0.5 -abs(position)  # reward is negative absolute distance to origin
        
        self.current_step += 1
        done = self.current_step >= self.max_steps
        truncated = False
        return self.state, reward, done, truncated, {}

    def render(self, mode='console'):
        if mode == 'console':
            print(f"Position: {self.state[0]}, Velocity: {self.state[1]}")

    def close(self):
        pass


import pandas as pd
from collections import defaultdict


class MetricsRecorder:
    def __init__(self):
        self.losses = {
            "total_losses": [],
            "policy_losses": [],
            "entropy_losses": [],
            "value_losses": [],
        }
        self.actions = {"action_means": [], "action_stds": []}
        self.episode_rewards = {
            "episode_reward_mins": [],
            "episode_reward_means": [],
            "episode_reward_maxs": [],
        }
        self.learning = defaultdict(list)

    def record_losses(self, total_loss, policy_loss, entropy_loss, value_loss):
        self.losses["total_losses"].append(total_loss)
        self.losses["policy_losses"].append(policy_loss)
        self.losses["entropy_losses"].append(entropy_loss)
        self.losses["value_losses"].append(value_loss)

    def record_actions(self, action_mean, action_std):
        self.actions["action_means"].append(action_mean)
        self.actions["action_stds"].append(action_std)

    def record_rewards(self, rewards):
        self.episode_rewards["episode_reward_mins"].append(min(rewards))
        self.episode_rewards["episode_reward_means"].append(sum(rewards) / len(rewards))
        self.episode_rewards["episode_reward_maxs"].append(max(rewards))

    def record_learning(self, lrs):
        for k, v in lrs.items():
            self.learning[k].append(v)

    def to_csv(self):
        import pandas as pd

        # Convert dictionaries to DataFrames
        df_losses = pd.DataFrame(self.losses)
        df_actions = pd.DataFrame(self.actions)
        df_rewards = pd.DataFrame(self.episode_rewards)
        df_learning = pd.DataFrame(self.learning)

        # Save to separate CSVs
        df_losses.to_csv("losses_metrics.csv", index=False)
        df_actions.to_csv("actions_metrics.csv", index=False)
        df_rewards.to_csv("rewards_metrics.csv", index=False)
        df_learning.to_csv("learning_metrics.csv", index=False)



import gym


class EnvironmentManager:
    def __init__(self, env_name, env_config:dict=None):
        self.env_name = env_name
        self.env_config = env_config

    def setup_env(self):
        if self.env_config:
            env = gym.make(self.env_name, config=self.env_config)
        else:
            env = gym.make(self.env_name)
        return env



# random_utils.py

import random
import numpy as np
import torch
import time


def set_seed(seed=None, use_torch=True):
    """
    Set the seed for all random number generators.
    If no seed is provided, sets a random seed based on current time.

    Parameters:
    - seed (int): The seed to set. If None, will set a random seed.
    - use_torch (bool): If True, will also set the seed for PyTorch.
    """
    if seed is None:
        seed = int((time.time() * 1e6) % 1e6)
    random.seed(seed)
    np.random.seed(seed)
    if use_torch:
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False


def get_seed():
    """
    Retrieve the current seed value for the Python random module.
    Note: This won't fetch seeds for numpy or torch.
    """
    return random.getstate()


def sample_action(prob_distribution):
    """
    Sample an action from a given probability distribution.

    Parameters:
    - prob_distribution (list or numpy array): The probability distribution over actions.

    Returns:
    - The sampled action index.
    """
    return np.random.choice(len(prob_distribution), p=prob_distribution)


def epsilon_greedy_action(values, epsilon=0.1):
    """
    Choose an action using epsilon-greedy strategy.

    Parameters:
    - values (list or numpy array): Estimated values or Q-values for each action.
    - epsilon (float): The probability of choosing a random action.

    Returns:
    - The chosen action index.
    """
    if random.random() < epsilon:
        return random.randint(0, len(values) - 1)
    else:
        return np.argmax(values)



from typing import Tuple
from pathlib import Path
from nanoppo.continuous_action_ppo import train
import ray
from ray import tune
import gym
from nanoppo.continuous_action_ppo import update_config
from nanoppo.random_utils import set_seed
from nanoppo.envs.point_mass2d import PointMass2DEnv

gym.envs.register("PointMass2D-v0", entry_point=PointMass2DEnv)

#set_seed(SEED)

# Define the hyperparameter search space
mountaincar_search_space = {
    "project": "tune_continuous_action_ppo",
    "env_name": "MountainCarContinuous-v0",
    "policy_lr": tune.loguniform(1e-6, 1e-3),
    "value_lr": tune.loguniform(1e-6, 1e-3),
    "weight_decay": tune.loguniform(1e-6, 1e-3),
    "scheduler": tune.choice([None, "exponential", "cosine"]),
    # "cosine_T_max": tune.choice([50, 100, 200, 500]),
    "sgd_iters": tune.choice([2, 5, 10, 20, 30]),
    # "scale_states": tune.choice(["standard", "minmax"]),
    #    "rescaling_rewards": True,
    #   "scale_states": tune.choice(["by_sample", "by_env", None]),
    #    "init_type": tune.choice(['he', 'xavier']),
    "batch_size": tune.choice([64, 128, 256, 512]),
    # "l1_loss": tune.choice([True, False]),
    "clip_param": tune.uniform(0.1, 0.3),
    "max_grad_norm": tune.uniform(0.5, 1.0),
    "vf_coef": tune.uniform(0.5, 1.0),
    "entropy_coef": tune.loguniform(1e-6, 1e-3),
    "tau": tune.loguniform(0.9, 1),
}

pendulum_search_space = {
    "project": "tune_continuous_action_ppo",
    "seed": tune.choice(range(1000)),
    "env_name": "Pendulum-v1",
    "policy_lr": tune.loguniform(1e-5, 1e-2),
    "value_lr": tune.loguniform(1e-5, 1e-2),
    "weight_decay": tune.loguniform(1e-4, 1e-2),
    # "cosine_T_max": tune.choice([ 200, 500]),
    "sgd_iters": tune.choice([2, 3, 5, 10]),
    # "scheduler": tune.choice([None, "exponential", "cosine"]),
    "scale_states": tune.choice(["standard", "minmax", "robust", "quantile"]),
    # "scheduler": None,
    # "scale_states": "robust",
    "use_gae": tune.choice([True, False]),
    "init_type": tune.choice(["he", "xavier"]),
    # "init_type": "xavier",
    # "batch_size": tune.choice([64, 128, 256, 512]),
    "batch_size": tune.choice([64, 128, 256, 512]),
    # "l1_loss": False,
    "clip_param": tune.loguniform(0.2, 0.5),
    "max_grad_norm": tune.uniform(0.5, 1),
    "vf_coef": tune.uniform(0.1, 2),
    "entropy_coef": tune.loguniform(1e-7, 1e-3),
    "tau": tune.loguniform(0.9, 1),
}

pointmass2d_search_space = {
    "project": "tune_continuous_action_ppo",
    #"seed": tune.choice(range(1000)),
    "env_name": "PointMass2D-v0",
    "policy_lr": tune.loguniform(1e-5, 1e-3),
    "value_lr": tune.loguniform(1e-5, 1e-3),
    "weight_decay": tune.loguniform(1e-4, 1e-2),
    # "cosine_T_max": tune.choice([ 200, 500]),
    "sgd_iters": tune.choice([2, 3, 5, 10]),
    # "scheduler": tune.choice([None, "exponential", "cosine"]),
    #"scale_states": tune.choice(["standard", "minmax", "robust", "quantile"]),
    # "scheduler": None,
    # "scale_states": "robust",
    #"use_gae": tune.choice([True, False]),
    "use_gae": True,
    #"init_type": tune.choice(["default", "he", "xavier"]),
    # "init_type": "xavier",
    # "batch_size": tune.choice([64, 128, 256, 512]),
    "batch_size": tune.choice([64, 128, 256, 512]),
    # "l1_loss": False,
    #"clip_param": tune.uniform(0.2, 0.4),
    #"max_grad_norm": tune.uniform(0.5, 0.7),
    "vf_coef": tune.uniform(0.1, 1),
    "entropy_coef": tune.loguniform(1e-7, 1e-3),
    #"tau": tune.loguniform(0.9, 1),
}


def training_function(search_config):
    #set_seed(search_config.pop("seed"))
    tune_config = update_config(search_config)
    tune_config.update({"verbose": False, "wandb_log": False})
    tune_config.update({'report_func': lambda **kwargs: tune.report(**kwargs)})

    policy, value, average_reward, total_iters = train(
        **tune_config
    )
    print("train", "average reward", average_reward, "total iter", total_iters)


stop_criteria = {"training_iteration": 50, "mean_reward": 90}

if __name__ == "__main__":
    ray.init(num_cpus=10, num_gpus=0, local_mode=False)
    #search_space = pendulum_search_space
    # search_space =  mountaincar_search_space
    # search_space = weight_search_space
    search_space = pointmass2d_search_space
    # Run the hyperparameter search with Tune
    analysis = tune.run(
        training_function,
        config=search_space,
        num_samples=50,  # Number of different configurations to try
        stop=stop_criteria,  # Stop after 10 iterations,
        metric="mean_reward",  # Compare trials by their mean reward
        progress_reporter=tune.CLIReporter(
            max_progress_rows=10,
            max_error_rows=10,
            metric_columns=["mean_reward", "training_iteration"],
        ),
        verbose=1,
        fail_fast=False
    )

    # Print the best configuration and its performance
    # best_config = analysis.get_best_config(metric="mean_reward", mode="max", scope="last-10-avg")
    # print(f"Best config: {best_config}")
    # print(f"Best performance: {analysis.get_best_trial(metric='mean_reward', mode='max', scope='last-10-avg')}")

    # Sort trials based on their performance
    sorted_trials = sorted(
        analysis.trials,
        key=lambda trial: trial.last_result.get("mean_reward", 0),
        reverse=True,
    )

    # Print the top 5 configurations and their performances
    for idx, trial in enumerate(sorted_trials[:5]):
        print(f"Rank {idx + 1}:")
        print(f"Config: {trial.config}")
        print(f"Mean Reward: {trial.last_result['mean_reward']}\n")



import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.optim.lr_scheduler import CosineAnnealingLR, ExponentialLR
import torch.optim as optim
import gym
import numpy as np
import wandb
from tqdm import tqdm
import random
import os
import glob
from pathlib import Path

import click
from copy import deepcopy
from nanoppo.reward_scaler import RewardScaler
from nanoppo.reward_shaper import (
    RewardShaper,
    MountainCarDirectionalRewardShaper,
    TDRewardShaper,
)
from nanoppo.rollout_buffer import RolloutBuffer
from nanoppo.state_scaler import StateScaler
from nanoppo.metrics_recorder import MetricsRecorder
from nanoppo.policy.network import PolicyNetwork, ValueNetwork
from nanoppo.random_utils import set_seed
from nanoppo.envs.point_mass1d import PointMass1DEnv 
from nanoppo.envs.point_mass2d import PointMass2DEnv 
from nanoppo.ppo_utils import compute_gae, compute_returns_and_advantages_without_gae
from nanoppo.normalizer import Normalizer
from time import time
import warnings

# Suppress the specific warning
warnings.filterwarnings(
    "ignore",
    message="Could not parse CUBLAS_WORKSPACE_CONFIG, using default workspace size of 8519680 bytes.",
)
gym.register("PointMass1D-v0", entry_point=PointMass1DEnv, max_episode_steps=200)
gym.register("PointMass2D-v0", entry_point=PointMass2DEnv, max_episode_steps=200)
# SEED = 153 # Set a random seed for reproducibility MountainviewCar 43 Pendulum 153
# set_seed(SEED)


def save_checkpoint(policy, value, optimizer, epoch, checkpoint_path):
    # Create checkpoint directory if it does not exist
    if not os.path.exists(checkpoint_path):
        os.makedirs(checkpoint_path)
    checkpoint = {
        "epoch": epoch,
        "policy_state_dict": policy.state_dict(),
        "value_state_dict": value.state_dict(),
        "optimizer_state_dict": optimizer.state_dict(),
    }
    torch.save(checkpoint, f"{checkpoint_path}/checkpoint_epoch{epoch}.pt")


def load_checkpoint(policy, value, optimizer, checkpoint_path, epoch=None):
    # Find the latest checkpoint file
    if epoch is None:
        checkpoint_files = sorted(glob.glob(f"{checkpoint_path}/checkpoint_epoch*.pt"))
        if len(checkpoint_files) == 0:
            raise ValueError("No checkpoint found in the specified directory.")
        checkpoint_file = checkpoint_files[-1]
    else:
        checkpoint_file = f"{checkpoint_path}/checkpoint_epoch{epoch}.pt"

    checkpoint = torch.load(checkpoint_file)
    policy.load_state_dict(checkpoint["policy_state_dict"])
    value.load_state_dict(checkpoint["value_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    epoch = checkpoint["epoch"]
    return epoch


def setup_env(env_name, env_config):
    if env_config:
        env = gym.make(env_name, config=env_config)
    else:
        env = gym.make(env_name)
    return env


def setup_networks(
    env,
    optimizer_config,
    hidden_size,
    init_type,
    device,
):
    if isinstance(env.observation_space, gym.spaces.Dict):
        observation_space = env.observation_space.spaces["obs"]
    else:
        observation_space = env.observation_space
    if isinstance(env.action_space, gym.spaces.Box):
        action_dim = env.action_space.shape[0]
    else:
        action_dim = env.action_space.n

    policy = PolicyNetwork(
        state_dim=observation_space.shape[0], action_dim=action_dim, n_latent_var=hidden_size, init_type=init_type
    ).to(device)
    value = ValueNetwork(
        observation_space.shape[0], hidden_size=hidden_size, init_type=init_type
    ).to(device)

    policy_lr = optimizer_config["policy_lr"]  # Add this to your config
    value_lr = optimizer_config["value_lr"]  # Add this to your config

    optimizer = optim.Adam(
        [
            {"params": policy.parameters(), "lr": policy_lr},
            {"params": value.parameters(), "lr": value_lr},
        ],
        betas=(optimizer_config["beta1"], optimizer_config["beta2"]),
        eps=optimizer_config["epsilon"],
        weight_decay=optimizer_config["weight_decay"],
    )

    if optimizer_config["scheduler"] is None:
        scheduler = None
    elif optimizer_config["scheduler"] == "exponential":
        scheduler = ExponentialLR(
            optimizer, gamma=optimizer_config["exponential_gamma"]
        )
    elif optimizer_config["scheduler"] == "cosine":
        # scheduler = CosineAnnealingLR(optimizer, T_max=epochs*sgd_iters)
        scheduler = CosineAnnealingLR(optimizer, T_max=optimizer_config["cosine_T_max"])
    else:
        raise ValueError(f"Scheduler {optimizer_config['scheduler']} not recognized.")
    return policy, value, optimizer, scheduler


def get_progress_iterator(last_epoch, epochs, verbose):
    if verbose:
        progress_iterator = tqdm(range(last_epoch, last_epoch + epochs))
    else:
        progress_iterator = range(last_epoch, last_epoch + epochs)
    return progress_iterator


def log_rewards(rewards):
    # Log the rewards during training
    wandb.log(
        {
            "Reward/Min": min(rewards),
            "Reward/Mean": sum(rewards) / len(rewards),
            "Reward/Max": max(rewards),
        }
    )


def rescale_action(predicted_action, action_min, action_max):
    rescaled_action =  action_min + (predicted_action - (-1)) / 2 * (action_max - action_min)
    return rescaled_action


def rollout_with_step(
    policy,
    value,
    env,
    device,
    rollout_buffer: RolloutBuffer,
    state_scaler: StateScaler,
    state_normalizer: Normalizer,
    reward_shaper: RewardShaper,
    reward_scaler: RewardScaler,
    wandb_log: bool,
    debug: bool,
):
    total_steps = 0
    while True:
        steps = 0
        state, info = env.reset()
        if isinstance(state, dict):
            state = state["obs"]
        if state_normalizer:
            state_normalizer.observe(state)
            scaled_state = state_normalizer.normalize(state)
        else:
            scaled_state = state_scaler.scale_state(state)
        done = False
        truncated = False
        accumulated_rewards = 0

        while (not done) and (not truncated):
            action, log_prob, action_mean, action_std = select_action(
                policy,
                scaled_state,
                device,
                env.action_space.low[0],
                env.action_space.high[0],
            )
            if wandb_log:
                wandb.log(
                    {
                        f"Policy/Action{i}_Mean": action_mean
                        for i, action_mean in enumerate(action_mean.numpy().tolist())
                    }
                )
                wandb.log(
                    {
                        f"Policy/Action{i}_Std": action_std
                        for i, action_std in enumerate(action_std.numpy().tolist())
                    }
                )

            #action = rescale_action(action, env.action_space.low[0], env.action_space.high[0])
            action = action.numpy()
            log_prob = log_prob.numpy()

            next_state, reward, done, truncated, info = env.step(action)
            if isinstance(next_state, dict):
                next_state = next_state["obs"]

            if state_scaler is None:
                scaled_next_state = next_state
            else:
                #scaled_next_state = state_scaler.scale_state(next_state)
                state_normalizer.observe(next_state)
                scaled_next_state = state_normalizer.normalize(next_state)

            scaled_state = scaled_next_state
            accumulated_rewards += reward
            # Reshape rewards
            if reward_shaper is None:
                reshaped_reward = reward
            else:
                reshaped_reward = reward_shaper.reshape([reward], [state], [next_state])[0]
            # Scale rewards
            if reward_scaler is None:
                scaled_reward = reshaped_reward
                # click.secho("Warning: Reward scaling is not applied.", fg="yellow", err=True)
            else:
                scaled_reward = reward_scaler.scale_rewards([reshaped_reward])[0]
            rollout_buffer.push(
                scaled_state,
                action.squeeze(),
                log_prob,
                scaled_reward,
                scaled_next_state,
                done,
            )
            total_steps += 1
            steps += 1
            if done or truncated:
                total_rewards = accumulated_rewards
                accumulated_rewards = 0
                if debug and total_steps % 1000 == 0:
                    print(
                        "total steps",
                        total_steps,
                        "steps",
                        steps,
                        "total_rewards[done|truncated]",
                        total_rewards,
                        "done",
                        done,
                        "truncated",
                        truncated,
                        "reward",
                        reward,
                        "reshaped_reward",
                        reshaped_reward,
                        "scaled_reward",
                        scaled_reward,
                    )
                yield total_rewards, steps, total_steps
            else:
                yield None, steps, total_steps



def surrogate(policy, old_probs, states, actions, advs, clip_param, entropy_coef):
    # Policy loss
    dist = policy(states)
    new_probs = dist.log_prob(actions).sum(-1)
    ratio = torch.exp(new_probs - old_probs)  # Importance sampling ratio
    surr1 = ratio * advs
    surr2 = (
        torch.clamp(ratio, 1 - clip_param, 1 + clip_param) * advs
    )  # Trust region clipping
    entropy = dist.entropy().mean()  # Compute the mean entropy of the distribution
    return (
        -torch.min(surr1, surr2).mean(),
        -entropy_coef * entropy,
    )  # Add the entropy term to the policy loss


def compute_value_loss(value, states, returns, l1_loss):
    # Compute value loss
    v_pred = value(states).squeeze()
    v_target = returns.squeeze()
    value_loss = (
        F.smooth_l1_loss(v_pred, v_target) if l1_loss else F.mse_loss(v_pred, v_target)
    )
    return value_loss


def select_action(policy: PolicyNetwork, state, device, action_min, action_max):
    state = torch.from_numpy(state).float().to(device)
    dist = policy(state)
    action = dist.sample()
    # If action space is continuous, compute the log_prob of the action, sum(-1) to sum over all dimensions
    log_prob = dist.log_prob(action).sum(-1)

    # Extract mean and std of the action distribution
    action_mean = dist.mean
    action_std = dist.stddev

    # action = torch.tanh(action)  # Pass the sampled action through the tanh activation function
    # action = action + torch.normal(mean=torch.zeros_like(action), std=action_std)  # Add noise to the action
    action = action.clamp(
        action_min, action_max
    )  # Clip the action to the valid range of the action space
    return (
        action.cpu().detach(),
        log_prob.cpu().detach(),
        action_mean.cpu().detach(),
        action_std.cpu().detach(),
    )

def train_networks(
    iter_num,
    policy,
    value,
    optimizer,
    scheduler,
    rollout_buffer,
    device,
    batch_size,
    sgd_iters,
    gamma,
    clip_param,
    vf_coef,
    entropy_coef,
    max_grad_norm,
    use_gae,
    tau,
    l1_loss,
    wandb_log,
    metrics_recorder: MetricsRecorder,
    verbose
):
    assert (
        len(rollout_buffer) >= batch_size
    ), f"Rollout buffer length {len(rollout_buffer)} is less than batch size {batch_size}"
    for sgd_iter in range(sgd_iters):
        # Sample from the rollout buffer
        (
            batch_states,
            batch_actions,
            batch_probs,
            batch_rewards,
            batch_next_states,
            batch_dones,
        ) = rollout_buffer.sample(batch_size, device=device, randomize=False)

        # Compute Advantage and Returns
        returns = []
        advs = []
        # Compute returns and advantages from rollout buffer samples out of order
        with torch.no_grad():
            if use_gae:
                # Compute Advantage using GAE and Returns
                values = value(batch_states).squeeze().tolist()
                next_value = value(batch_next_states[-1]).item()
                masks = [1 - done.item() for done in batch_dones]
                returns = compute_gae(
                    next_value, batch_rewards, masks, values, gamma, tau
                )
                advs = [ret - val for ret, val in zip(returns, values)]
            else:
                raise NotImplementedError
                # Compute Advantage and Returns without GAE
                returns, advs = compute_returns_and_advantages_without_gae(
                    batch_rewards, batch_states, batch_next_states, batch_dones, value, gamma
                )

        returns = torch.tensor(returns, dtype=torch.float32).to(device)
        advs = torch.tensor(advs, dtype=torch.float32).to(device)
        advs = (advs - advs.mean()) / (advs.std() + 1e-10)

        # Create mini-batches
        num_samples = len(batch_rewards)
        assert num_samples == batch_size
        num_batches = num_samples // batch_size
        assert num_batches == 1
        
        policy_loss, entropy_loss = surrogate(
            policy,
            old_probs=batch_probs,
            states=batch_states,
            actions=batch_actions,
            advs=advs,
            clip_param=clip_param,
            entropy_coef=entropy_coef,
        )
        """
        # clear the list of activation norms for each epoch
        activation_norms = []
        def hook(module, input, output):
            if isinstance(output, Tuple):
                for o in output:
                    activation_norms.append(o.norm().item())
            else:
                activation_norms.append(output.norm().item())
        # register activation norm hook
        # add forward hook to each layer of the value network
        hooks = []
        for module in value.modules():
            hooks.append(module.register_forward_hook(hook))
        """

        value_loss = compute_value_loss(value, batch_states, returns, l1_loss)

        # Compute total loss and update parameters
        total_loss = policy_loss + entropy_loss + vf_coef * value_loss

        optimizer.zero_grad()
        total_loss.backward()

        # Clip the gradients to avoid exploding gradients
        policy_grad_norm = nn.utils.clip_grad_norm_(policy.parameters(), max_grad_norm)
        value_grad_norm = nn.utils.clip_grad_norm_(value.parameters(), max_grad_norm)

        # compute activation norm
        # remove the forward hooks
        """
        for hook in hooks:
            hook.remove()
        # compute the mean activation norm for the value network
        activation_norm = sum(activation_norms) / len(activation_norms)
        """
    
        # 1. Loss Magnitude Monitoring
        abs_policy_loss = torch.abs(policy_loss).item()
        abs_value_loss = torch.abs(value_loss).item()
        abs_total_loss = torch.abs(total_loss).item()
    
        # 2. Gradient Magnitude Monitoring
        policy_gradient_magnitude = torch.norm(
            torch.cat([p.grad.view(-1) for p in policy.parameters()])
        ).item()
        value_gradient_magnitude = torch.norm(
            torch.cat([p.grad.view(-1) for p in value.parameters()])
        ).item()
    
        # 3. Learning Rate Monitoring
        policy_lr = optimizer.param_groups[0]['lr']
        value_lr = optimizer.param_groups[1]['lr']
    
        # 4. Parameter Magnitude
        policy_param_magnitude = torch.norm(
            torch.cat([p.data.view(-1) for p in policy.parameters()])
        ).item()
        value_param_magnitude = torch.norm(
            torch.cat([p.data.view(-1) for p in value.parameters()])
        ).item()


        optimizer.step()
        if scheduler is not None:
            scheduler.step()
    
        if wandb_log:
            # ... [your existing wandb.log code here]
    
            # Log additional metrics
            wandb.log({
                'Magnitude/AbsPolicyLoss': abs_policy_loss,
                'Magnitude/AbsValueLoss': abs_value_loss,
                'Magnitude/AbsTotalLoss': abs_total_loss,
                'Magnitude/PolicyGradient': policy_gradient_magnitude,
                'Magnitude/ValueGradient': value_gradient_magnitude,
                'Magnitude/PolicyParameters': policy_param_magnitude,
                'Magnitude/ValueParameters': value_param_magnitude,
                'LR/Policy': policy_lr,
                'LR/Value': value_lr
            })

        if wandb_log:
            # Log the losses and gradients to WandB
            wandb.log(
                {
                    "iteration": iter_num,
                    "Loss/Total": total_loss.item(),
                    "Loss/Policy": policy_loss.item(),
                    "Loss/Entropy": entropy_loss.item(),
                    "Loss/Value": value_loss.item(),
                    "Loss/Coef_Value": vf_coef * value_loss.item(),
                }
            )
            # wandb.log({"Gradients/PolicyNet": wandb.Histogram(policy.fc1.weight.grad.detach().cpu().numpy())})
            wandb.log(
                {
                    "Policy/Gradient_Norm": policy_grad_norm,
                    "Value/Gradient_Norm": value_grad_norm,
                    # "Value/Activation_Norm": activation_norm
                }
            )
            # wandb.log({"Gradients/ValueNet": wandb.Histogram(value.fc1.weight.grad.detach().cpu().numpy())})
            log_std_values = policy.action_log_std(batch_states).detach().cpu().numpy()
            average_log_std = log_std_values.mean()
            wandb.log({"Policy/Log_Std": average_log_std})
        # log the learning rate to wandb
        lrs = {}
        for i, param_group in enumerate(optimizer.param_groups):
            lr = param_group["lr"]
            lrs["learning_rate_{}".format(i)] = lr
            if wandb_log:
                wandb.log({"LR/LearningRate_{}".format(i): lr})

        if metrics_recorder:
            metrics_recorder.record_losses(
                total_loss.item(),
                policy_loss.item(),
                entropy_loss.item(),
                value_loss.item(),
            )
            metrics_recorder.record_learning(lrs)

        iter_num += 1
        
        if verbose > 1 and iter_num % 1000 == 0:
            print('iter', iter_num, 'sgd iter', sgd_iter, 'loss', total_loss.item(), 'policy loss', policy_loss.item(), 'entropy loss', entropy_loss.item(), 'value loss', value_loss.item(), 'policy grad norm', policy_grad_norm, 'value grad norm', value_grad_norm, 'policy lr', policy_lr, 'value lr', value_lr, 'policy param magnitude', policy_param_magnitude, 'value param magnitude', value_param_magnitude, 'policy gradient magnitude', policy_gradient_magnitude, 'value gradient magnitude', value_gradient_magnitude)
    return (
        policy,
        value,
        iter_num,
    )


def train(
    env_name: str,
    env_config: dict,
    shape_reward: RewardShaper,
    rescaling_rewards: bool,
    scale_states: str,
    epochs: int,
    batch_size: int,
    sgd_iters: int,
    gamma: float,
    optimizer_config: dict,
    hidden_size: int,
    init_type: str,
    clip_param: float,
    vf_coef: float,
    entropy_coef: float,
    max_grad_norm: float,
    use_gae: bool,
    tau: float,
    l1_loss: bool,
    wandb_log: bool,
    verbose: int,
    checkpoint_interval: int = -1,
    checkpoint_path: str = None,
    resume_training: bool = False,
    resume_epoch: bool = None,
    report_func: callable = None,
    project: str = "continuous-action-ppo",
    seed: int = None,
):
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    # Initialize WandB
    if wandb_log:
        config = locals().copy()
        keys_to_delete = [
            "checkpoint_interval",
            "checkpoint_path",
            "resume_training",
            "resume_epoch",
            "report_func",
        ]
        [config.pop(key, None) for key in keys_to_delete]
        wandb.init(project=project, name=env_name, config=config)
    metrics_recorder = MetricsRecorder()

    # Set up environment and neural networks
    env = setup_env(env_name, env_config)
    policy, value, optimizer, scheduler = setup_networks(
        env,
        optimizer_config,
        hidden_size=hidden_size,
        init_type=init_type,
        device=device,
    )
    rollout_buffer = RolloutBuffer(batch_size)

    if rescaling_rewards:
        reward_scaler = RewardScaler()
    else:
        reward_scaler = None
        click.secho("No reward scaling", fg="red", err=True)
    
    state_normalizer = None
    state_scaler = None
    if scale_states == 'default':
        state_normalizer = Normalizer(env.observation_space.shape[0])
    elif scale_states in ['env', 'standard', 'minmax', 'robust', 'quantile']:
        state_scaler = StateScaler(env, sample_size=10000, scale_type=scale_states)
    else:
        raise ValueError(f"State scaler {scale_states} not recognized.")

    if resume_training:
        last_epoch = load_checkpoint(
            policy, value, optimizer, checkpoint_path, epoch=resume_epoch
        )
    else:
        last_epoch = 0

    if shape_reward is None:
        reward_shaper = None
    elif shape_reward == TDRewardShaper:
        reward_shaper = TDRewardShaper(model=value, device=device)
    else:
        reward_shaper = shape_reward()

    # Set up rollout generator
    rollout = rollout_with_step(
        policy,
        value,
        env,
        device,
        rollout_buffer,
        state_scaler,
        state_normalizer,
        reward_shaper,
        reward_scaler,
        wandb_log,
        debug=True if verbose > 1 else False,
    )
    # Set up training loop
    episode_rewards = []
    episode_steps = []
    total_iters = epochs * sgd_iters
    train_iters = 0
    rollout_steps = 0
    average_reward = -np.inf
    start = time()
    for epoch in get_progress_iterator(last_epoch, epochs, verbose):
        policy.eval()
        value.eval()
        rollout_buffer.clear()  # clear the rollout buffer, all data is from the current policy
        for r in range(batch_size):
            total_rewards, steps, rollout_steps = next(rollout)
            if total_rewards is not None:
                episode_steps.append(steps)
                episode_rewards.append(total_rewards)
        policy.train()
        value.train()
        _, _, train_iters = train_networks(
            train_iters,
            policy,
            value,
            optimizer,
            scheduler,
            rollout_buffer,
            device,
            batch_size,
            sgd_iters,
            gamma,
            clip_param,
            vf_coef,
            entropy_coef,
            max_grad_norm,
            use_gae=use_gae,
            tau=tau,
            l1_loss=l1_loss,
            wandb_log=wandb_log,
            metrics_recorder=metrics_recorder,
            verbose=verbose
        )
        if checkpoint_interval > 0 and ((epoch + 1) % checkpoint_interval) == 0:
            save_checkpoint(policy, value, optimizer, epoch, checkpoint_path)

        # Average of last 10 episodes or all episodes if less than 10 episodes are available
        
        if len(episode_rewards) >= 20:
            average_reward = sum(episode_rewards[-20:]) / len(episode_rewards[-20:])
            if wandb_log:
                log_rewards(episode_rewards[-20:])
            if metrics_recorder:
                metrics_recorder.record_rewards(episode_rewards[-20:])
            if report_func:
                report_func(mean_reward=average_reward)  # Reporting the reward

        if verbose > 0 and (epoch + 1) % 10 == 0:
            print(
                "epoch",
                epoch + 1,
                "average reward",
                average_reward,
                "train epochs",
                epoch - last_epoch + 1,
                "train iters",
                train_iters,
                "rollout episodes",
                len(episode_rewards),
                "rollout steps",
                rollout_steps,
            )
    end = time()
    print("Training time: ", round((end - start) / 60, 2), "minutes")

    metrics_recorder.to_csv()
    if wandb_log:
        wandb.finish()
    print(
        "Training complete",
        "average reward",
        average_reward,
        "total iters",
        total_iters,
    )
    return policy, value, average_reward, train_iters


config = {
    "project": "continuous-action-ppo",
    "env_name": "MountainCarContinuous-v0",
    "env_config": None,
    "seed": None,
    "epochs": 500,
    "rescaling_rewards": False,
    "shape_reward": None,  # [None, SubClass of RewardReshaper]
    "scale_states": "default",  # ["default", "env", "standard", "minmax", "robust", "quantile"]:
    "init_type": "default",  # ["default", "xavier", "he"]
    "use_gae":True, 
    "gamma": 0.99,
    "tau": 0.95,
    "l1_loss": False,
    "hidden_size": 128,
    "batch_size": 256,
    "sgd_iters": 4,
    "vf_coef": 0.5,
    "clip_param": 0.2,
    "max_grad_norm": 2,
    "entropy_coef": 1e-2,
    "wandb_log": True,
    "verbose": 2,
    "checkpoint_interval": 100,
}

optimizer_config = {
    "policy_lr": 5 * 1e-4,
    "value_lr": 5 * 1e-4,
    "beta1": 0.9,
    "beta2": 0.999,
    "epsilon": 1e-8,
    "weight_decay": 1e-4,
    "scheduler": None,  # None, exponential, cosine
    "exponential_gamma": 0.9995,  # for exponential scheduler only
    "cosine_T_max": config["epochs"] * config["sgd_iters"],  # for cosine scheduler only
}


def update_config(aconfig):
    o = deepcopy(optimizer_config)
    c = deepcopy(config)
    for key in aconfig:
        if key in o:
            o[key] = aconfig[key]
        elif key in c:
            c[key] = aconfig[key]
        else:
            raise ValueError(f"Key {key} not recognized.")
    c.update(
        {
            "optimizer_config": o,
            "checkpoint_path": str(Path("checkpoints") / c["project"] / c["env_name"]),
        }
    )
    return c

def train_env(env_name):
    if env_name == "MountainCarContinuous-v0":
        best_config = {
            "env_name": "MountainCarContinuous-v0",
        }
    elif env_name == "Pendulum-v1":
        # best_config = {'epochs':100}
        best_config = {}
    elif env_name == "LunarLanderContinuous-v2":
        pass
    elif env_name == "BipedalWalker-v3":
        pass
    elif env_name == "PointMass1D-v0":
        best_config = {
            "env_name": "PointMass1D-v0",
        }
    elif env_name == "PointMass2D-v0":
        best_config = {'project': 'tune_continuous_action_ppo', 'env_name': 'PointMass2D-v0', 'policy_lr': 6.965606472889233e-05, 'value_lr': 5.337021407693641e-05, 'weight_decay': 0.005615599464308048, 'sgd_iters': 2, 'use_gae': True, 'batch_size': 256, 'vf_coef': 0.4407130713151379, 'entropy_coef': 4.796586082455764e-05}

    best_config["env_name"] = env_name
    train_config = update_config(best_config)
    print("train config", train_config)
    set_seed(train_config.pop("seed", None))
    policy, value, average_reward, total_iters = train(**train_config)
    print("train", "average reward", average_reward, "total iters", total_iters)

@click.command()
@click.option("--env_name", default="PointMass2D-v0", type=click.Choice(["PointMass1D-v0", "PointMass2D-v0", "MountainCarContinuous-v0", "Pendulum-v1"]))
def cli(env_name):
    train_env(env_name=env_name)

if __name__ == "__main__":
    cli()



from copy import deepcopy
from pathlib import Path


def update_config(config, optimizer_config, aconfig):
    o = deepcopy(optimizer_config)
    c = deepcopy(config)
    for key in aconfig:
        if key in o:
            o[key] = aconfig[key]
        elif key in c:
            c[key] = aconfig[key]
        else:
            raise ValueError(f"Key {key} not recognized.")
    c.update(
        {
            "optimizer_config": o,
            "checkpoint_path": str(Path("checkpoints") / c["project"] / c["env_name"]),
        }
    )
    return c



def eval(
    policy, env_name, state_scaler, env_config={}, device="cpu", epochs=20, render=False
):
    policy.eval()
    policy.to(device)
    if env_config:
        env = gym.make(env_name, config=env_config)
    else:
        env = gym.make(env_name)
    env.seed(SEED)
    action_min = env.action_space.low[0]
    action_max = env.action_space.high[0]
    episode_rewards = []
    actions = []
    for epoch in range(epochs):
        state, info = env.reset()
        if isinstance(state, dict):
            state = state["obs"]
        scaled_state = state_scaler.scale_state(state)
        done = False
        total_reward = 0
        steps = 0
        while not done:
            with torch.no_grad():
                action, _ = select_action(
                    policy, scaled_state, device, action_min, action_max
                )
            action = action.numpy()
            actions.append(action)
            next_state, reward, done, truncated, info = env.step(action)
            done = done or truncated
            if isinstance(next_state, dict):
                next_state = next_state["obs"]
            scaled_next_state = state_scaler.scale_state(next_state)
            if render:
                env.render()
            scaled_state = scaled_next_state
            total_reward += reward
            steps += 1
        else:
            episode_rewards.append(total_reward)
    policy.train()
    return episode_rewards, actions
